/*
 * Copyright (c) 2017 Digital Bazaar, Inc. All rights reserved.
 */
'use strict';

const JsonLdError = require('./JsonLdError');

var _require = require('./types');

const _isArray = _require.isArray,
      _isObject = _require.isObject,
      _isString = _require.isString;

var _require2 = require('./graphTypes');

const _isList = _require2.isList,
      _isValue = _require2.isValue,
      _isSimpleGraph = _require2.isSimpleGraph,
      _isSubjectReference = _require2.isSubjectReference;

var _require3 = require('./context');

const _expandIri = _require3.expandIri,
      _getContextValue = _require3.getContextValue,
      _isKeyword = _require3.isKeyword;

var _require4 = require('./url');

const _removeBase = _require4.removeBase;

var _require5 = require('./util');

const _addValue = _require5.addValue,
      _compareShortestLeast = _require5.compareShortestLeast;


const api = {};
module.exports = api;

/**
 * Recursively compacts an element using the given active context. All values
 * must be in expanded form before this method is called.
 *
 * @param activeCtx the active context to use.
 * @param activeProperty the compacted property associated with the element
 *          to compact, null for none.
 * @param element the element to compact.
 * @param options the compaction options.
 * @param compactionMap the compaction map to use.
 *
 * @return the compacted value.
 */
api.compact = (_ref) => {
  let activeCtx = _ref.activeCtx;
  var _ref$activeProperty = _ref.activeProperty;
  let activeProperty = _ref$activeProperty === undefined ? null : _ref$activeProperty,
      element = _ref.element;
  var _ref$options = _ref.options;
  let options = _ref$options === undefined ? {} : _ref$options;
  var _ref$compactionMap = _ref.compactionMap;
  let compactionMap = _ref$compactionMap === undefined ? () => undefined : _ref$compactionMap;

  // recursively compact array
  if (_isArray(element)) {
    let rval = [];
    for (let i = 0; i < element.length; ++i) {
      // compact, dropping any null values unless custom mapped
      let compacted = api.compact({
        activeCtx,
        activeProperty,
        element: element[i],
        options,
        compactionMap
      });
      if (compacted === null) {
        // TODO: use `await` to support async
        compacted = compactionMap({
          unmappedValue: element[i],
          activeCtx,
          activeProperty,
          parent: element,
          index: i,
          options
        });
        if (compacted === undefined) {
          continue;
        }
      }
      rval.push(compacted);
    }
    if (options.compactArrays && rval.length === 1) {
      // use single element if no container is specified
      const container = _getContextValue(activeCtx, activeProperty, '@container');
      if (container === null) {
        rval = rval[0];
      }
    }
    return rval;
  }

  // recursively compact object
  if (_isObject(element)) {
    if (options.link && '@id' in element && element['@id'] in options.link) {
      // check for a linked element to reuse
      const linked = options.link[element['@id']];
      for (let i = 0; i < linked.length; ++i) {
        if (linked[i].expanded === element) {
          return linked[i].compacted;
        }
      }
    }

    // do value compaction on @values and subject references
    if (_isValue(element) || _isSubjectReference(element)) {
      const rval = api.compactValue({ activeCtx, activeProperty, value: element });
      if (options.link && _isSubjectReference(element)) {
        // store linked element
        if (!(element['@id'] in options.link)) {
          options.link[element['@id']] = [];
        }
        options.link[element['@id']].push({ expanded: element, compacted: rval });
      }
      return rval;
    }

    // FIXME: avoid misuse of active property as an expanded property?
    const insideReverse = activeProperty === '@reverse';

    const rval = {};

    if (options.link && '@id' in element) {
      // store linked element
      if (!(element['@id'] in options.link)) {
        options.link[element['@id']] = [];
      }
      options.link[element['@id']].push({ expanded: element, compacted: rval });
    }

    // process element keys in order
    const keys = Object.keys(element).sort();
    for (let ki = 0; ki < keys.length; ++ki) {
      const expandedProperty = keys[ki];
      const expandedValue = element[expandedProperty];

      // compact @id and @type(s)
      if (expandedProperty === '@id' || expandedProperty === '@type') {
        let compactedValue;

        // compact single @id
        if (_isString(expandedValue)) {
          compactedValue = api.compactIri({
            activeCtx,
            iri: expandedValue,
            relativeTo: { vocab: expandedProperty === '@type' }
          });
        } else {
          // expanded value must be a @type array
          compactedValue = [];
          for (let vi = 0; vi < expandedValue.length; ++vi) {
            compactedValue.push(api.compactIri({ activeCtx, iri: expandedValue[vi], relativeTo: { vocab: true } }));
          }
        }

        // use keyword alias and add value
        const alias = api.compactIri({ activeCtx, iri: expandedProperty });
        const isArray = _isArray(compactedValue) && expandedValue.length === 0;
        _addValue(rval, alias, compactedValue, { propertyIsArray: isArray });
        continue;
      }

      // handle @reverse
      if (expandedProperty === '@reverse') {
        // recursively compact expanded value
        const compactedValue = api.compact({
          activeCtx,
          activeProperty: '@reverse',
          element: expandedValue,
          options,
          compactionMap
        });

        // handle double-reversed properties
        for (let compactedProperty in compactedValue) {
          if (activeCtx.mappings[compactedProperty] && activeCtx.mappings[compactedProperty].reverse) {
            const value = compactedValue[compactedProperty];
            const container = [].concat(_getContextValue(activeCtx, compactedProperty, '@container'));
            const useArray = container.includes('@set') || !options.compactArrays;
            _addValue(rval, compactedProperty, value, { propertyIsArray: useArray });
            delete compactedValue[compactedProperty];
          }
        }

        if (Object.keys(compactedValue).length > 0) {
          // use keyword alias and add value
          const alias = api.compactIri({ activeCtx, iri: expandedProperty });
          _addValue(rval, alias, compactedValue);
        }

        continue;
      }

      // handle @index property
      if (expandedProperty === '@index') {
        // drop @index if inside an @index container
        const container = _getContextValue(activeCtx, activeProperty, '@container');
        if (container === '@index') {
          continue;
        }

        // use keyword alias and add value
        const alias = api.compactIri({ activeCtx, iri: expandedProperty });
        _addValue(rval, alias, expandedValue);
        continue;
      }

      // skip array processing for keywords that aren't @graph or @list
      if (expandedProperty !== '@graph' && expandedProperty !== '@list' && _isKeyword(expandedProperty)) {
        // use keyword alias and add value as is
        const alias = api.compactIri({ activeCtx, iri: expandedProperty });
        _addValue(rval, alias, expandedValue);
        continue;
      }

      // Note: expanded value must be an array due to expansion algorithm.
      if (!_isArray(expandedValue)) {
        throw new JsonLdError('JSON-LD expansion error; expanded value must be an array.', 'jsonld.SyntaxError');
      }

      // preserve empty arrays
      if (expandedValue.length === 0) {
        const itemActiveProperty = api.compactIri({
          activeCtx,
          iri: expandedProperty,
          value: expandedValue,
          relativeTo: { vocab: true },
          reverse: insideReverse
        });
        _addValue(rval, itemActiveProperty, expandedValue, { propertyIsArray: true });
      }

      // recusively process array values
      for (let vi = 0; vi < expandedValue.length; ++vi) {
        const expandedItem = expandedValue[vi];

        // compact property and get container type
        const itemActiveProperty = api.compactIri({
          activeCtx,
          iri: expandedProperty,
          value: expandedItem,
          relativeTo: { vocab: true },
          reverse: insideReverse
        });
        const container = [].concat(_getContextValue(activeCtx, itemActiveProperty, '@container'));

        // get simple @graph or @list value if appropriate
        const isSimpleGraph = _isSimpleGraph(expandedItem);
        const isList = _isList(expandedItem);
        let inner;
        if (isList) {
          inner = expandedItem['@list'];
        } else if (isSimpleGraph) {
          inner = expandedItem['@graph'];
        }

        // recursively compact expanded item
        let compactedItem = api.compact({
          activeCtx,
          activeProperty: itemActiveProperty,
          element: isList || isSimpleGraph ? inner : expandedItem,
          options,
          compactionMap
        });

        // handle @list
        if (isList) {
          // ensure @list value is an array
          if (!_isArray(compactedItem)) {
            compactedItem = [compactedItem];
          }

          if (!container.includes('@list')) {
            // wrap using @list alias
            compactedItem = {
              [api.compactIri({ activeCtx, iri: '@list' })]: compactedItem
            };

            // include @index from expanded @list, if any
            if ('@index' in expandedItem) {
              compactedItem[api.compactIri({ activeCtx, iri: '@index' })] = expandedItem['@index'];
            }
          } else if (itemActiveProperty in rval) {
            // can't use @list container for more than 1 list
            throw new JsonLdError('JSON-LD compact error; property has a "@list" @container ' + 'rule but there is more than a single @list that matches ' + 'the compacted term in the document. Compaction might mix ' + 'unwanted items into the list.', 'jsonld.SyntaxError', { code: 'compaction to list of lists' });
          }
        }

        // handle simple @graph
        if (isSimpleGraph && !container.includes('@graph')) {
          // wrap using @graph alias
          compactedItem = {
            [api.compactIri({ activeCtx, iri: '@graph' })]: compactedItem
          };

          // include @index from expanded @graph, if any
          if ('@index' in expandedItem) {
            compactedItem[api.compactIri({ activeCtx, iri: '@index' })] = expandedItem['@index'];
          }
        }

        // handle language and index maps
        if (container.includes('@language') || container.includes('@index')) {
          // get or create the map object
          let mapObject;
          if (itemActiveProperty in rval) {
            mapObject = rval[itemActiveProperty];
          } else {
            rval[itemActiveProperty] = mapObject = {};
          }

          // if container is a language map, simplify compacted value to
          // a simple string
          if (container.includes('@language') && _isValue(compactedItem)) {
            compactedItem = compactedItem['@value'];
          }

          // add compact value to map object using key from expanded value
          // based on the container type
          _addValue(mapObject, expandedItem[container], compactedItem);
        } else {
          // use an array if: compactArrays flag is false,
          // @container is @set or @list , value is an empty
          // array, or key is @graph
          const isArray = !options.compactArrays || container.includes('@set') || container.includes('@list') || _isArray(compactedItem) && compactedItem.length === 0 || expandedProperty === '@list' || expandedProperty === '@graph';

          // add compact value
          _addValue(rval, itemActiveProperty, compactedItem, { propertyIsArray: isArray });
        }
      }
    }

    return rval;
  }

  // only primitives remain which are already compact
  return element;
};

/**
 * Compacts an IRI or keyword into a term or prefix if it can be. If the
 * IRI has an associated value it may be passed.
 *
 * @param activeCtx the active context to use.
 * @param iri the IRI to compact.
 * @param value the value to check or null.
 * @param relativeTo options for how to compact IRIs:
 *          vocab: true to split after @vocab, false not to.
 * @param reverse true if a reverse property is being compacted, false if not.
 *
 * @return the compacted term, prefix, keyword alias, or the original IRI.
 */
api.compactIri = (_ref2) => {
  let activeCtx = _ref2.activeCtx,
      iri = _ref2.iri;
  var _ref2$value = _ref2.value;
  let value = _ref2$value === undefined ? null : _ref2$value;
  var _ref2$relativeTo = _ref2.relativeTo;
  let relativeTo = _ref2$relativeTo === undefined ? { vocab: false } : _ref2$relativeTo;
  var _ref2$reverse = _ref2.reverse;
  let reverse = _ref2$reverse === undefined ? false : _ref2$reverse;

  // can't compact null
  if (iri === null) {
    return iri;
  }

  const inverseCtx = activeCtx.getInverse();

  // if term is a keyword, it can only be compacted to a simple alias
  if (_isKeyword(iri)) {
    if (iri in inverseCtx) {
      return inverseCtx[iri]['@none']['@type']['@none'];
    }
    return iri;
  }

  // use inverse context to pick a term if iri is relative to vocab
  if (relativeTo.vocab && iri in inverseCtx) {
    const defaultLanguage = activeCtx['@language'] || '@none';

    // prefer @index if available in value
    const containers = [];
    if (_isObject(value) && '@index' in value) {
      containers.push('@index');
    }

    // prefer `['@graph', '@set']` and then `@graph` if value is a simple graph
    // TODO: support `@graphId`?
    if (_isSimpleGraph(value)) {
      containers.push('@graph@set');
      containers.push('@graph');
    }

    // defaults for term selection based on type/language
    let typeOrLanguage = '@language';
    let typeOrLanguageValue = '@null';

    if (reverse) {
      typeOrLanguage = '@type';
      typeOrLanguageValue = '@reverse';
      containers.push('@set');
    } else if (_isList(value)) {
      // choose the most specific term that works for all elements in @list
      // only select @list containers if @index is NOT in value
      if (!('@index' in value)) {
        containers.push('@list');
      }
      const list = value['@list'];
      let commonLanguage = list.length === 0 ? defaultLanguage : null;
      let commonType = null;
      for (let i = 0; i < list.length; ++i) {
        const item = list[i];
        let itemLanguage = '@none';
        let itemType = '@none';
        if (_isValue(item)) {
          if ('@language' in item) {
            itemLanguage = item['@language'];
          } else if ('@type' in item) {
            itemType = item['@type'];
          } else {
            // plain literal
            itemLanguage = '@null';
          }
        } else {
          itemType = '@id';
        }
        if (commonLanguage === null) {
          commonLanguage = itemLanguage;
        } else if (itemLanguage !== commonLanguage && _isValue(item)) {
          commonLanguage = '@none';
        }
        if (commonType === null) {
          commonType = itemType;
        } else if (itemType !== commonType) {
          commonType = '@none';
        }
        // there are different languages and types in the list, so choose
        // the most generic term, no need to keep iterating the list
        if (commonLanguage === '@none' && commonType === '@none') {
          break;
        }
      }
      commonLanguage = commonLanguage || '@none';
      commonType = commonType || '@none';
      if (commonType !== '@none') {
        typeOrLanguage = '@type';
        typeOrLanguageValue = commonType;
      } else {
        typeOrLanguageValue = commonLanguage;
      }
    } else {
      if (_isValue(value)) {
        if ('@language' in value && !('@index' in value)) {
          containers.push('@language');
          typeOrLanguageValue = value['@language'];
        } else if ('@type' in value) {
          typeOrLanguage = '@type';
          typeOrLanguageValue = value['@type'];
        }
      } else {
        typeOrLanguage = '@type';
        typeOrLanguageValue = '@id';
      }
      containers.push('@set');
    }

    // do term selection
    containers.push('@none');
    const term = _selectTerm(activeCtx, iri, value, containers, typeOrLanguage, typeOrLanguageValue);
    if (term !== null) {
      return term;
    }
  }

  // no term match, use @vocab if available
  if (relativeTo.vocab) {
    if ('@vocab' in activeCtx) {
      // determine if vocab is a prefix of the iri
      const vocab = activeCtx['@vocab'];
      if (iri.indexOf(vocab) === 0 && iri !== vocab) {
        // use suffix as relative iri if it is not a term in the active context
        const suffix = iri.substr(vocab.length);
        if (!(suffix in activeCtx.mappings)) {
          return suffix;
        }
      }
    }
  }

  // no term or @vocab match, check for possible CURIEs
  let choice = null;
  // TODO: make FastCurieMap a class with a method to do this lookup
  const partialMatches = [];
  let iriMap = activeCtx.fastCurieMap;
  // check for partial matches of against `iri`, which means look until
  // iri.length - 1, not full length
  const maxPartialLength = iri.length - 1;
  for (let i = 0; i < maxPartialLength && iri[i] in iriMap; ++i) {
    iriMap = iriMap[iri[i]];
    if ('' in iriMap) {
      partialMatches.push(iriMap[''][0]);
    }
  }
  // check partial matches in reverse order to prefer longest ones first
  for (let i = partialMatches.length - 1; i >= 0; --i) {
    const entry = partialMatches[i];
    const terms = entry.terms;
    for (let ti = 0; ti < terms.length; ++ti) {
      // a CURIE is usable if:
      // 1. it has no mapping, OR
      // 2. value is null, which means we're not compacting an @value, AND
      //   the mapping matches the IRI
      const curie = terms[ti] + ':' + iri.substr(entry.iri.length);
      const isUsableCurie = !(curie in activeCtx.mappings) || value === null && activeCtx.mappings[curie]['@id'] === iri;

      // select curie if it is shorter or the same length but lexicographically
      // less than the current choice
      if (isUsableCurie && (choice === null || _compareShortestLeast(curie, choice) < 0)) {
        choice = curie;
      }
    }
  }

  // return chosen curie
  if (choice !== null) {
    return choice;
  }

  // compact IRI relative to base
  if (!relativeTo.vocab) {
    return _removeBase(activeCtx['@base'], iri);
  }

  // return IRI as is
  return iri;
};

/**
 * Performs value compaction on an object with '@value' or '@id' as the only
 * property.
 *
 * @param activeCtx the active context.
 * @param activeProperty the active property that points to the value.
 * @param value the value to compact.
 *
 * @return the compaction result.
 */
api.compactValue = (_ref3) => {
  let activeCtx = _ref3.activeCtx,
      activeProperty = _ref3.activeProperty,
      value = _ref3.value;

  // value is a @value
  if (_isValue(value)) {
    // get context rules
    const type = _getContextValue(activeCtx, activeProperty, '@type');
    const language = _getContextValue(activeCtx, activeProperty, '@language');
    const container = _getContextValue(activeCtx, activeProperty, '@container');

    // whether or not the value has an @index that must be preserved
    const preserveIndex = '@index' in value && container !== '@index';

    // if there's no @index to preserve ...
    if (!preserveIndex) {
      // matching @type or @language specified in context, compact value
      if (value['@type'] === type || value['@language'] === language) {
        return value['@value'];
      }
    }

    // return just the value of @value if all are true:
    // 1. @value is the only key or @index isn't being preserved
    // 2. there is no default language or @value is not a string or
    //   the key has a mapping with a null @language
    const keyCount = Object.keys(value).length;
    const isValueOnlyKey = keyCount === 1 || keyCount === 2 && '@index' in value && !preserveIndex;
    const hasDefaultLanguage = '@language' in activeCtx;
    const isValueString = _isString(value['@value']);
    const hasNullMapping = activeCtx.mappings[activeProperty] && activeCtx.mappings[activeProperty]['@language'] === null;
    if (isValueOnlyKey && (!hasDefaultLanguage || !isValueString || hasNullMapping)) {
      return value['@value'];
    }

    const rval = {};

    // preserve @index
    if (preserveIndex) {
      rval[api.compactIri({ activeCtx, iri: '@index' })] = value['@index'];
    }

    if ('@type' in value) {
      // compact @type IRI
      rval[api.compactIri({ activeCtx, iri: '@type' })] = api.compactIri({ activeCtx, iri: value['@type'], relativeTo: { vocab: true } });
    } else if ('@language' in value) {
      // alias @language
      rval[api.compactIri({ activeCtx, iri: '@language' })] = value['@language'];
    }

    // alias @value
    rval[api.compactIri({ activeCtx, iri: '@value' })] = value['@value'];

    return rval;
  }

  // value is a subject reference
  const expandedProperty = _expandIri(activeCtx, activeProperty, { vocab: true });
  const type = _getContextValue(activeCtx, activeProperty, '@type');
  const compacted = api.compactIri({ activeCtx, iri: value['@id'], relativeTo: { vocab: type === '@vocab' } });

  // compact to scalar
  if (type === '@id' || type === '@vocab' || expandedProperty === '@graph') {
    return compacted;
  }

  return {
    [api.compactIri({ activeCtx, iri: '@id' })]: compacted
  };
};

/**
 * Removes the @preserve keywords as the last step of the compaction
 * algorithm when it is running on framed output.
 *
 * @param ctx the active context used to compact the input.
 * @param input the framed, compacted output.
 * @param options the compaction options used.
 *
 * @return the resulting output.
 */
api.removePreserve = (ctx, input, options) => {
  // recurse through arrays
  if (_isArray(input)) {
    const output = [];
    for (let i = 0; i < input.length; ++i) {
      const result = api.removePreserve(ctx, input[i], options);
      // drop nulls from arrays
      if (result !== null) {
        output.push(result);
      }
    }
    input = output;
  } else if (_isObject(input)) {
    // remove @preserve
    if ('@preserve' in input) {
      if (input['@preserve'] === '@null') {
        return null;
      }
      return input['@preserve'];
    }

    // skip @values
    if (_isValue(input)) {
      return input;
    }

    // recurse through @lists
    if (_isList(input)) {
      input['@list'] = api.removePreserve(ctx, input['@list'], options);
      return input;
    }

    // handle in-memory linked nodes
    const idAlias = api.compactIri({ activeCtx: ctx, iri: '@id' });
    if (idAlias in input) {
      const id = input[idAlias];
      if (id in options.link) {
        const idx = options.link[id].indexOf(input);
        if (idx !== -1) {
          // already visited
          return options.link[id][idx];
        }
        // prevent circular visitation
        options.link[id].push(input);
      } else {
        // prevent circular visitation
        options.link[id] = [input];
      }
    }

    // recurse through properties
    for (let prop in input) {
      let result = api.removePreserve(ctx, input[prop], options);
      const container = _getContextValue(ctx, prop, '@container');
      if (options.compactArrays && _isArray(result) && result.length === 1 && container === null) {
        result = result[0];
      }
      input[prop] = result;
    }
  }
  return input;
};

/**
 * Picks the preferred compaction term from the given inverse context entry.
 *
 * @param activeCtx the active context.
 * @param iri the IRI to pick the term for.
 * @param value the value to pick the term for.
 * @param containers the preferred containers.
 * @param typeOrLanguage either '@type' or '@language'.
 * @param typeOrLanguageValue the preferred value for '@type' or '@language'.
 *
 * @return the preferred term.
 */
function _selectTerm(activeCtx, iri, value, containers, typeOrLanguage, typeOrLanguageValue) {
  if (typeOrLanguageValue === null) {
    typeOrLanguageValue = '@null';
  }

  // preferences for the value of @type or @language
  const prefs = [];

  // determine prefs for @id based on whether or not value compacts to a term
  if ((typeOrLanguageValue === '@id' || typeOrLanguageValue === '@reverse') && _isSubjectReference(value)) {
    // prefer @reverse first
    if (typeOrLanguageValue === '@reverse') {
      prefs.push('@reverse');
    }
    // try to compact value to a term
    const term = api.compactIri({ activeCtx, iri: value['@id'], relativeTo: { vocab: true } });
    if (term in activeCtx.mappings && activeCtx.mappings[term] && activeCtx.mappings[term]['@id'] === value['@id']) {
      // prefer @vocab
      prefs.push.apply(prefs, ['@vocab', '@id']);
    } else {
      // prefer @id
      prefs.push.apply(prefs, ['@id', '@vocab']);
    }
  } else {
    prefs.push(typeOrLanguageValue);
  }
  prefs.push('@none');

  const containerMap = activeCtx.inverse[iri];
  for (let ci = 0; ci < containers.length; ++ci) {
    // if container not available in the map, continue
    const container = containers[ci];
    if (!(container in containerMap)) {
      continue;
    }

    const typeOrLanguageValueMap = containerMap[container][typeOrLanguage];
    for (let pi = 0; pi < prefs.length; ++pi) {
      // if type/language option not available in the map, continue
      const pref = prefs[pi];
      if (!(pref in typeOrLanguageValueMap)) {
        continue;
      }

      // select term
      return typeOrLanguageValueMap[pref];
    }
  }

  return null;
}